#include <stdio.h>  /* debuging*/
#include <string.h>
#include <assert.h>
#include <glib.h>
#include <cUnit.h>

#include <sys/types.h>
#include <sys/stat.h> 

#include <fcntl.h>


#include "gios.h"

#define MAX_ARG_LEN 255

static char** _argv = NULL;

#ifndef __FUNCTION__
#define __FUNCTION__ "__FUNCTION__"
#endif

#define CHK_ERR(err) if (err!=NULL) {\
    g_warning (" chk_err: %s(%d)/%d/%s\n", __FUNCTION__, __LINE__, err->code, err->message);\
    cUnit_assert (err==NULL);\
}

#define READ  0
#define WRITE 1

typedef enum 
{
   PARENT=0,
   TST_PROCESS_NEW,
   TST_PIPE,
   TST_DUP,
   TST_POLL,
   TST_SOCKET,
   TST_ACCEPT,
   TST_LISTEN,
   TST_CONNECT,
   TST_OPEN,
   MODE_MAX
} TstType;


 
static
TstType
get_cmd (int argc, char** argv)
{
   if (argc==1 
       ||  g_ascii_strcasecmp (argv[1], "--TestRunner")==0)
   {
      return PARENT;
   }
   else 
   { 
      return atoi (argv[1]);
   }
}

static 
int
get_arg (int argc, char** argv, int idx)
{
   return atoi (argv[2+idx]);
}


gchar*
tst_socket (gpointer data)
{
   { 
      GError* err=NULL; 
      gint sd = g_socket (PF_INET, SOCK_STREAM, 0, &err);
      cUnit_assert (err==NULL);
   }
   /* unix specific */
	/*  {
      GError* err;
      gint sd = g_socket (PF_UNIX, SOCK_STREAM, AF_INET, &err);
      cUnit_assert (err!=NULL);
      g_error_free (err);
   }*/
   return NULL;
}


gchar*
tst_process_new__parent (gpointer data)
{
   GError* err = NULL;
   gint proc      = 0;
   gint exit_code = 0;
   
   char cmd[255+1];
   char arg0[255+1];

   char* args[] = {NULL, NULL, NULL, NULL};
   args[0] = _argv[0];
   args[1] = cmd;
   args[2] = arg0;
   
   /* create new process, no stdio specified */
   for (exit_code=0; exit_code<128; exit_code+=4)
   {
      g_snprintf (cmd, 255, "%d", TST_PROCESS_NEW);
      g_snprintf (arg0, 255,  "%d", exit_code);

      cUnit_assert ((proc=g_process_new (args, -1, -1, -1, 0, &err))
		    !=-1); 
      {
	 gint stat=0;
	 gint ret=g_process_wait (proc, &stat, 100*1000, NULL);
	 cUnit_assert (ret<0 || stat==exit_code);
      }
      g_process_free (proc, NULL);
   }

      /* create new process, no stdio specified */
   for (exit_code=0; exit_code<128; exit_code+=4)
   {
      g_snprintf (cmd, 255, "%d", TST_PROCESS_NEW);
      g_snprintf (arg0, 255,  "%d", exit_code);

      cUnit_assert ((proc=g_process_new (args, -1, -1, -1, 0, &err))
		    !=-1); 
      {
	 gint stat=0;
	 gint ret=g_process_wait (proc, &stat, 100*1000, NULL);
	 cUnit_assert (ret<0 || stat==exit_code);
      }
      g_process_free (proc, NULL);
   }

   /* create new process, only stdin specified */
   for (exit_code=0; exit_code<128; exit_code+=4)
   {
      g_snprintf (cmd, 255, "%d", TST_PROCESS_NEW);
      g_snprintf (arg0, 255,  "%d", exit_code);

      cUnit_assert ((proc=g_process_new (args, G_STDIN, -1, -1, 0, &err))
		    !=-1); 
      {
	 gint stat=0;
	 gint ret=g_process_wait (proc, &stat, 100*1000, NULL);
	 cUnit_assert (ret<0 || stat==exit_code);
      }
      g_process_free (proc, NULL);
   }

   /* create new process, only stdout specified */
   for (exit_code=0; exit_code<128; exit_code+=4)
   {
      g_snprintf (cmd, 255, "%d", TST_PROCESS_NEW);
      g_snprintf (arg0, 255,  "%d", exit_code);

      cUnit_assert ((proc=g_process_new (args, -1, G_STDOUT, -1, 0, &err))
		    !=-1); 

      {
	 gint stat=0;
	 gint ret=g_process_wait (proc, &stat, 100*1000, &err);
	 cUnit_assert (ret<0 || stat==exit_code);
      }
      g_process_free (proc, NULL);
   }

   /* create new process, only stderr specified */
   for (exit_code=0; exit_code<128; exit_code+=4)
   {
      g_snprintf (cmd, 255, "%d", TST_PROCESS_NEW);
      g_snprintf (arg0, 255,  "%d", exit_code);

      cUnit_assert ((proc=g_process_new (args, -1, -1, G_STDERR, 0, &err))
		    !=-1); 
    
      {
	 gint stat=0;
	 gint ret=g_process_wait (proc, &stat, 100*1000, NULL);
	 cUnit_assert (ret<0 || stat==exit_code);
      }
      g_process_free (proc, NULL);
   }

   /* create new process, only stdout specified */
   for (exit_code=0; exit_code<128; exit_code+=4)
   {
      g_snprintf (cmd, 255, "%d", TST_PROCESS_NEW);
      g_snprintf (arg0, 255,  "%d", exit_code);

      cUnit_assert ((proc=g_process_new (args, G_STDIN, G_STDOUT, G_STDERR, 0, &err))
		    !=-1); 
      {
	 gint stat=0;
	 gint ret=g_process_wait (proc, &stat, 100*1000, NULL);
	 cUnit_assert (ret<0 || stat==exit_code);
      }
      g_process_free (proc, NULL);
   }

   return NULL;
}

void
tst_process_new__child (int argc, char** argv)
{
   gint exit_code = get_arg(argc, argv, 0);
  g_sleep (10, NULL);
   exit (exit_code);
}

gchar*
tst_pipe__parent (gpointer data)
{
   GError* err = NULL;
   gint n=0;
   char c;
   gint  proc         = -1;
   gint  to_child_pipe[2]  = {-1, -1};
   gint  to_parent_pipe[2] = {-1, -1};

   char  arg_cmd [MAX_ARG_LEN];
   char* args[]={NULL,NULL,NULL,NULL,NULL,NULL,NULL};
  
   g_pipe (to_child_pipe,  &err); CHK_ERR (err); 
   g_pipe (to_parent_pipe, &err); CHK_ERR (err); 
 
   g_snprintf (arg_cmd, MAX_ARG_LEN, "%d", TST_PIPE);
   
   args[0] = _argv[0];
   args[1] =  arg_cmd;
   args[2] =  NULL;

   proc=g_process_new (args, 
		       to_child_pipe[READ], 
		       to_parent_pipe[WRITE], 
		       G_STDERR, 
		       0, 
		       &err);
   CHK_ERR (err); 
   cUnit_assert (proc>0);

   {
      gint i;
 	  	
      c=2;
	   /* test with file IO operation */
      for (i=0;i<333; ++i)
      {
	  n=g_write (to_child_pipe[WRITE], &c, sizeof(c), &err);
	  CHK_ERR (err); 
	  cUnit_assert (n==1);
	 
	  n=g_read(to_parent_pipe[READ], &c, sizeof(c), &err);

	  CHK_ERR (err); 
	  cUnit_assert (n==1);
      }
      
      // send termination signal to child
      c=0;
      g_write (to_child_pipe[WRITE], &c, sizeof(c), &err); CHK_ERR (err); 
   }	 

   /* finally close local endings of pipes */
   g_close (to_child_pipe[WRITE],  &err); CHK_ERR (err);
   g_close (to_child_pipe[READ],   &err); CHK_ERR (err);
   g_close (to_parent_pipe[WRITE], &err); CHK_ERR (err);
   g_close (to_parent_pipe[READ],  &err); CHK_ERR (err);

   // child must terminate with 0
   {
      gint stat=0;
      gint ret=g_process_wait (proc, &stat, 100*1000, &err);
      cUnit_assert (ret<0 || stat==0);
   }
   
   g_process_free (proc, NULL);

   return NULL;
}

void
tst_pipe__child (int argc, char** argv)
{
   GError* err=NULL;
   
   // duplicating must fail, because non_inherit_fd is private to
   // parent

   {
      char c;
      gint i=0;
      while (1)
      {
	 gint n = g_read (G_STDIN, &c, sizeof(c), &err);
	 if (n!=1)  /* terminate with error */
	    exit (1);

	 /* fprintf (stderr, "%d", c); */

	 if (c==0) 
	    exit (0);

	 n = g_write (G_STDOUT, &c, sizeof (c), &err);
	 if (n!=1)  /* terminate with error */
	    exit (1);
      }
   }
}

gchar*
tst_dup__parent (gpointer data)
{
   GError* err = NULL;
   
   gint  c        = 1;
   gint  proc     = -1;
   gint  to_parent_pipe[2] = {-1, -1};
   gint  from_child = -1;
   char  arg_cmd [MAX_ARG_LEN+1];
   char  arg_from_child [MAX_ARG_LEN+1];
   char* args[]={NULL,NULL,NULL,NULL,NULL,NULL,NULL};
  
   g_pipe (to_parent_pipe, &err);                          CHK_ERR (err); 
   from_child = g_dup (to_parent_pipe[READ], FALSE, &err); CHK_ERR (err);
   g_close (to_parent_pipe[READ], &err);                   CHK_ERR (err);
   
   g_snprintf (arg_cmd, MAX_ARG_LEN, "%d", TST_DUP);
   g_snprintf (arg_from_child, MAX_ARG_LEN, "%d", from_child);
   
   args[0] = _argv[0];
   args[1] =  arg_cmd;
   args[2] =  arg_from_child;
   args[3] =  NULL;

   proc=g_process_new (args, 
		       -1, 
		       G_STDOUT, 
		       G_STDERR, 
		       0, &err);
   CHK_ERR (err); 
   cUnit_assert (proc>0);
   
   g_write (to_parent_pipe[WRITE], &c, 1, &err); CHK_ERR (err); 
   g_close (to_parent_pipe[WRITE], &err);        CHK_ERR (err); 

   g_close (from_child, &err);

   {
      gint stat=0;
      gint ret=g_process_wait (proc, &stat, 10*1000, &err); /* FIXME, error occures here */
      if (ret < 0)
      {
	 // problem on win32: child did terminate before calling waitpid_with_timeout()
	 g_error_free (err); 
	 err=NULL;
      }
      else
      {
	 cUnit_assert (stat==0);
      }
      g_process_free (proc, NULL);
   }
   {
      /* duplicate stdin file descriptor  */
      gint stdin_cpy = g_dup (G_STDIN, FALSE, &err);    CHK_ERR (err); 
      g_close (stdin_cpy, &err);                        CHK_ERR (err);
      
      /* closing prev closed Fd should fail */
      cUnit_assert (g_dup (stdin_cpy, FALSE, &err) < 0 ); 
      cUnit_assert (err!=NULL);                          
      g_error_free (err);
   }

   return NULL;
}

void
tst_dup__child (int argc, char** argv)
{
   gint non_inherit_fd = atoi (argv[2]);
   
   // duplicating must fail, because non_inherit_fd is private to
   // parent

   char c = 1;
   gint n = 0;

   g_sleep (2*1000, NULL);

   /*   g_warning ("try to read from fd=%d\n", non_inherit_fd); */
   n=g_read(non_inherit_fd, &c, 1, NULL);
   if (n>0)
   {
      // terminate with error
      exit (1);
   }
   else
   {
      // terminate with success
      exit (0);
   }
}

#define POLL_CHUNK_SIZE (2<<16)

gchar*
tst_poll__parent (gpointer data)
{
   GError* err = NULL;
   
   gint  child          = -1;
   gint  to_child_pipe[2]  = {-1, -1};
   gint  to_parent_pipe[2] = {-1, -1};
   
   gint  to_child   = -1;
   gint  from_child = -1;

   char  arg_cmd [MAX_ARG_LEN+1];
   char* args[]={NULL,NULL,NULL,NULL,NULL,NULL,NULL};
  
   g_pipe (to_child_pipe,  &err); CHK_ERR (err); 
   g_pipe (to_parent_pipe, &err); CHK_ERR (err); 

   to_child   = g_dup (to_child_pipe[WRITE], FALSE, &err); CHK_ERR (err);
   from_child = g_dup (to_parent_pipe[READ], FALSE, &err); CHK_ERR (err);
   g_close (to_child_pipe[WRITE], &err);                   CHK_ERR (err);
   g_close (to_parent_pipe[READ], &err);                   CHK_ERR (err);
   
   g_snprintf (arg_cmd, MAX_ARG_LEN, "%d", TST_POLL);
   
   args[0] = _argv[0];
   args[1] =  arg_cmd;
   args[2] =  NULL;

   child=g_process_new (args, 
			to_child_pipe[READ], 
			to_parent_pipe[WRITE], 
			G_STDERR, 
			0, 
			&err);
   CHK_ERR (err); 
   cUnit_assert (child>0);

   g_close (to_child_pipe[READ],   &err); CHK_ERR (err); 
   g_close (to_parent_pipe[WRITE], &err); CHK_ERR (err); 
   
   G_IO_NONBLOCKING (to_child,   &err); CHK_ERR (err);
   G_IO_NONBLOCKING (from_child, &err); CHK_ERR (err);

   {
      GPollFD ufds[2];
      char buf[POLL_CHUNK_SIZE];

      gint i;
      gint sent=0;
      gint recv=0;

      for (i=0; i<POLL_CHUNK_SIZE; ++i)
	 buf[i] = 'A';


      ufds[0].fd      = to_child;
      ufds[0].events  = G_IO_OUT;
      ufds[0].revents = 0;
      ufds[1].fd      = from_child;
      ufds[1].events  = G_IO_IN;
      ufds[1].revents = 0;

      while (sent!=sizeof(buf) && recv!=sizeof(buf))
      {
	 gint n=0;

	 // wait infinite time, until ready for read or write
	 gint nevents=g_poll (ufds, 2, -1, &err);

	 cUnit_assert (nevents>0);
	 CHK_ERR (err); 
	 
	 cUnit_assert ( ! (ufds[0].revents & G_IO_IN) );
	 cUnit_assert ( ! (ufds[0].revents & G_IO_ERR) );
	 cUnit_assert ( ! (ufds[0].revents & G_IO_HUP) );
	 cUnit_assert ( ! (ufds[0].revents & G_IO_NVAL) );
	 cUnit_assert ( ! (ufds[0].revents & G_IO_PRI) );

	 if (ufds[0].revents & G_IO_OUT)
	 {
	    sent+=g_write (to_child, 
			   &buf[sent], 
			   sizeof(buf)-sent, 
			   &err); 
	    CHK_ERR (err); 
	 }
	 
	 cUnit_assert ( ! (ufds[1].revents & G_IO_OUT));
	 cUnit_assert ( ! (ufds[1].revents & G_IO_PRI));
	 cUnit_assert ( ! (ufds[1].revents & G_IO_ERR));
	 cUnit_assert ( ! (ufds[1].revents & G_IO_HUP));
	 cUnit_assert ( ! (ufds[1].revents & G_IO_NVAL));

	 if (ufds[1].revents & G_IO_IN)
	 {
	    recv+=g_read (from_child, 
			   &buf[recv], 
			   sizeof(buf)-recv,
			   &err);
	    CHK_ERR (err); 
	 }
      }

      // verify characters
      for (i=0; i<sizeof(buf); ++i)
	 cUnit_assert (buf[i]=='A');
   }

   {
	   char c=0;
	   g_write (to_child, &c, sizeof(c), &err);  CHK_ERR (err); 
   }

   G_IO_BLOCKING (to_child, &err);   CHK_ERR (err);
   G_IO_BLOCKING (from_child, &err); CHK_ERR (err);


   g_close (to_child, &err);      CHK_ERR (err); 
   g_close (from_child, &err);    CHK_ERR (err); 

   // child must terminate with 0
   {
      gint stat=0;
      gint ret=g_process_wait (child, &stat, 10*1000, NULL);
      cUnit_assert (ret<0 || stat==0);
   }
   
   g_process_free (child, NULL);

   return NULL;
}

void
tst_poll__child (int argc, char** argv)
{
   // duplicating must fail, because non_inherit_fd is private to
   // parent

   {
      char buf;
      while (1)
      {
	     gint n=0;
	     gint sent=0;

		 /* child processes do no cUnit tests, therefore &err==NULL*/
	     n=g_read (G_STDIN, &buf, sizeof(buf), NULL); 
	     if (n<0)     /* parent closed connection, terminate with success */
	        exit (1);

		 if (buf==0) 
		 {
		     /* fprintf (stderr, ";"); */
			 exit (0);
		 }
		 else
		 {
		   /*  fprintf (stderr, ".");*/
		 }

		 /* child processes do no cUnit tests, therefore &err==NULL*/
	     sent=g_write (G_STDOUT, &buf, sizeof(buf), NULL); 	 
	  }
   }
}


gchar*
tst_open__parent (gpointer data)
{
   {
      GError* err  = NULL;
      gint    mode = G_FILE_IRUSR|G_FILE_IWUSR|G_FILE_IRGRP|G_FILE_IWGRP;
      gint    file = -1;
      gchar   filename[1024];
      gint    pid=g_random_int() % 8888; 
      gchar*  l1 = "1 1 1 1 1 1 1 1 1 1\n";
      gchar*  l2 = "2 2 2 2 2 2 2 2 2 2\n";
      gchar   lx[1024];
      gint    n=0;
      
      CHK_ERR (err);

      g_snprintf (filename, 1024, "tst_open__file%d.txt", pid);
      
      
      /* test create */
      file = g_open (filename, G_FILE_RDWR|G_FILE_CREAT|G_FILE_TRUNC, mode, &err); CHK_ERR (err);
      cUnit_assert (file!=-1);
      g_write (file, l1, strlen(l1), &err);   CHK_ERR (err);
      g_close (file, &err);  CHK_ERR (err);
      
      /* test append */
      file = g_open (filename, G_FILE_WRONLY|G_FILE_APPEND, 0, &err); CHK_ERR (err);
      cUnit_assert (file!=-1);
      g_write (file, l2, strlen(l2), &err);   CHK_ERR (err);
      g_close (file, &err);  CHK_ERR (err);
      
      /* start reading */
      file = g_open (filename, G_FILE_RDONLY, 0, &err); CHK_ERR (err);
      cUnit_assert (file!=-1);
      
      /* read fist line l1 */
      n = g_read (file, lx, strlen(l1),  &err);    CHK_ERR (err);
      cUnit_assert (n==strlen(l1));
      cUnit_assert (g_ascii_strncasecmp (lx, l1, strlen(l1))==0);
      
      /* read appended line l2 */
      n = g_read (file, lx, strlen(l2),  &err);    CHK_ERR (err);
      cUnit_assert (n==strlen(l2));
      cUnit_assert (g_ascii_strncasecmp (lx, l2, strlen(l2))==0);
      g_close (file, &err);                        CHK_ERR (err);

      /* remove file from directory */
      n = g_remove (filename, &err);               CHK_ERR (err);
      cUnit_assert (n!=-1);
   }

   {
      GError* err=NULL;
      gchar*  argv[10];
      gchar   tst_id[256];
      gchar   filename[256];
      gchar   c=1;

      gint mode = G_FILE_IRUSR|G_FILE_IWUSR|G_FILE_IRGRP|G_FILE_IWGRP;
      gint file=-1;
      gint child1=0;
      gint child2=0;
      gint stat=-1;
      gint to_child1[2]={-1,-1};
      gint to_child2[2]={-1,-1};
      gint pid=g_getpid(&err); CHK_ERR (err);
      
      g_pipe (to_child1, &err);  CHK_ERR (err);
      g_pipe (to_child2, &err);  CHK_ERR (err);

      g_snprintf (tst_id, 255, "%d", TST_OPEN);
      g_snprintf (filename, 255, "tst_open__file%d.txt", pid);
      
      file=g_open (filename, G_FILE_RDWR|G_FILE_CREAT, mode, &err); CHK_ERR (err);
      cUnit_assert (file!=-1);
      
      argv[0]=_argv[0];
      argv[1]=tst_id;
      argv[2]=filename;
      argv[3]=NULL;

      child1=g_process_new (argv, 
			    to_child1[READ], 
			    -1,
			    G_STDERR, 
			    0, 
			    &err);
      CHK_ERR (err); 
      cUnit_assert (child1>0);
      
      child2=g_process_new (argv, 
			    to_child2[READ], 
			    -1,
			    G_STDERR, 
			    0, 
			    &err);
      CHK_ERR (err); 
      cUnit_assert (child2>0);
      
      g_process_wait (child1, &stat, 10*1000, NULL);
      g_process_wait (child2, &stat, 10*1000, NULL); 
      
      /* signal to childs: open file*/

      g_write (to_child1[WRITE], &c, sizeof(c), &err); CHK_ERR (err);
      g_write (to_child2[WRITE], &c, sizeof(c), &err); CHK_ERR (err);
      
      /* sleep till childs are ready */
      g_sleep (50 /* msecs */, &err);  CHK_ERR (err);

      /* signal to childs: flood file*/
      g_write (to_child1[WRITE], &c, sizeof(c), &err); CHK_ERR (err);
      g_write (to_child2[WRITE], &c, sizeof(c), &err); CHK_ERR (err);
      
      /* sleep till childs are ready */
      g_sleep (50 /* msecs */, &err);  CHK_ERR (err);

      /* after at least 10 seconds both childs should terminate */
      g_process_wait (child1, &stat, 10*1000, NULL);
      g_process_wait (child2, &stat, 10*1000, NULL);

      g_close (file,         &err); CHK_ERR (err);

      /* open file and count lines */
      file=g_open (filename, G_FILE_RDONLY, mode, &err); CHK_ERR (err);
      
      {
	 gint cnt=0;
	 gchar c=0;
	 
	 while (g_read (file, &c, sizeof(c), NULL)==sizeof(c))
	    if (c=='\n') ++cnt;

	 cUnit_assert (cnt==2000);
      }
      g_close (file,         &err); CHK_ERR (err);

      /* remove file from directory */
      cUnit_assert (g_remove (filename, &err)!=-1);    CHK_ERR (err);

      g_close (to_child1[0], &err); CHK_ERR (err);
      g_close (to_child1[1], &err); CHK_ERR (err);
      g_close (to_child2[0], &err); CHK_ERR (err);
      g_close (to_child2[1], &err); CHK_ERR (err);

      g_process_free (child1, NULL);
      g_process_free (child2, NULL);
   }

   return NULL;
}


void
tst_open__child (gint argc, gchar* argv[])
{
   assert (argc==3);
   {
      GError* err=NULL;
      gint    mode = G_FILE_IRUSR|G_FILE_IWUSR|G_FILE_IRGRP|G_FILE_IWGRP;
      gint    file=-1;
      gint    i=0;
      gchar   c=0;
      gchar   mesg[256];
      gchar*  filename = argv[2];

      g_snprintf (mesg, sizeof(mesg)-1, "%d blahbla\n", g_getpid(NULL));
      
      assert (g_read (G_STDIN, &c, sizeof(c), NULL)==sizeof(c));  

      file=g_open (filename, G_FILE_RDWR|G_FILE_APPEND|G_FILE_CREAT, mode, &err); assert (err==NULL);
      assert (file!=-1);

      assert (g_read (G_STDIN, &c, sizeof(c), NULL)==sizeof(c));  

      for (i=0; i<1000; ++i)
      {
	 g_write(file, mesg, strlen(mesg), NULL);
	 g_sleep (10, NULL);
      }

      g_close (file, NULL);

      exit (0);
   }
}

/* ********* my_build_session ******** */
/* 
 This function registeres the unit tests to a cUnit_session. It will
 be called by cUnit_TestRunner () if --TestRunner is set on the
 commandline.
*/
cUnit_session *
my_build_session (void)
{
   cUnit_session *session;
   
   /* Get a new test session */
   session = cUnit_session_new ();
   
   /* register the function tst_make_currency as unit test
      'make_currency0' on the suite 'currency' */
   cUnit_register_test     (session, "tst_process_new",
			    tst_process_new__parent);
   
   cUnit_register_test     (session, "tst_pipe",
			    tst_pipe__parent);
   
   cUnit_register_test     (session, "tst_dup",
			    tst_dup__parent);
   
   cUnit_register_test     (session, "tst_poll",
			    tst_poll__parent);

   cUnit_register_test     (session, "tst_socket", 
 			    tst_socket); 

   cUnit_register_test     (session, "tst_open", 
 			    tst_open__parent); 
   /* ... and return the test session */
   return (session);
}



	 

gint
main (gint argc, gchar *argv[])
{
   _argv = argv;

   switch (get_cmd (argc, argv))
   {
     case PARENT:             
	 /* check for --TestRunner */
	    cUnit_TestRunner (argc, argv, my_build_session);
	    break;
     case TST_PROCESS_NEW: 
	    tst_process_new__child (argc, argv);
	    break;
     case TST_PIPE:
	    tst_pipe__child(argc, argv);
	    break;
     case TST_POLL:
	    tst_poll__child(argc, argv);
	    break;
     case TST_DUP:
	    tst_dup__child(argc, argv);
	    break;
     case TST_OPEN:
	    tst_open__child(argc, argv);
	    break;
     default:
	    assert( ! "unknown mode");
   }

   exit (0);
}

